 #ifndef OPERATORS_H
#define OPERATORS_H 1

#include <vector>
#include <cstdlib>
#include "densematrix.h"
#include "sparsematrix.h"

/* This header contains Matrix-Vector and Vector-Vector operations
 * for the matrix types
 *  - DenseMatrix
 *  - SparseMatrix
 * and the vector types
 *  - std::vector<double>
 */

//=====================================
// Operators for SparseMatrix
//=====================================

SparseMatrix
operator+(const SparseMatrix& A, const SparseMatrix& B){
    if(A.numRows()!=B.numRows() || A.numCols()!=B.numCols()){
        std::cerr << "Error: Matrix Dimensions do not fit" << std::endl;
        std::exit(1);
    }
    SparseMatrix C(A);
    for(std::map<std::pair<int,int>,double>::const_iterator it = B.getData().begin(); it != B.getData().end(); ++it){
        C.operator()((*it).first.first, (*it).first.second) += (*it).second;
    }
    
    return C;
}

SparseMatrix
operator-(const SparseMatrix& A, const SparseMatrix& B){
    if(A.numRows()!=B.numRows() || A.numCols()!=B.numCols()){
        std::cerr << "Error: Matrix Dimensions do not fit" << std::endl;
        std::exit(1);
    }
    SparseMatrix C(A);
    for(std::map<std::pair<int,int>,double>::const_iterator it = B.getData().begin(); it != B.getData().end(); ++it){
        C.operator()((*it).first.first, (*it).first.second) -= (*it).second;
    }
    
    return C;
}

SparseMatrix
operator*(const double alpha, const SparseMatrix& rhs){
    SparseMatrix ret(rhs);
    for(std::map<std::pair<int,int>,double>::const_iterator it = rhs.getData().begin(); it != rhs.getData().end(); ++it){
        ret.operator()((*it).first.first, (*it).first.second) *= alpha;
    }
    return ret;
}

//=====================================
// Operators for DenseeMatrix
//=====================================
DenseMatrix
operator*(const DenseMatrix& A, const DenseMatrix& B)
{
    if(A.numCols() != B.numRows()){
        std::cerr << "Error: Matrix Dimensions do not fit" << std::endl;
        std::exit(1);
    }
    DenseMatrix C(A.numRows(), B.numCols());
    for(int i = 0; i < A.numRows(); ++i){
        for(int j = 0; j < B.numCols(); ++j){
            for(int k = 0; k < A.numCols(); ++k){
                C(i,j) += A(i,k)*B(k,j);
            }
        }
    }
    return C;
}

//=====================================
// Operators for std::vector<double>
//=====================================

std::vector<double>
operator*(const DenseMatrix &m, const std::vector<double> &v){
    
    std::vector<double> ret(m.numRows());
    for(int i = 0; i< m.numRows(); ++i){
        
        ret[i] = 0.0;
        for(int j = 0; j < m.numCols(); ++j){
            ret[i] += m(i,j)*v[j];
        }
    }
    return ret;
}

std::vector<double>
operator-(const std::vector<double> &v1, const std::vector<double> &v2){
    std::vector<double> ret(v1.size());
    for(size_t i = 0; i < v1.size(); ++i){
        ret[i] = v1[i] - v2[i];
    }
    return ret;
}

std::vector<double>
operator+(const std::vector<double> &v1, const std::vector<double> &v2){
    std::vector<double> ret(v1.size());
    for(size_t i = 0; i < v1.size(); ++i){
        ret[i] = v1[i] + v2[i];
    }
    return ret;
}



std::vector<double>
operator*(const double scalar, const std::vector<double> &v){
    //std::cout << "v.size() = " << v.size() << std::endl;
    std::vector<double> ret(v.size());
    for(size_t i = 0; i < v.size(); ++i){
        ret[i] = v[i] * scalar;
    }
    return ret;
}

double
operator*(const std::vector<double> &v1, const std::vector<double> &v2){
    double ret=0.0;
    for(size_t i = 0; i < v1.size(); ++i){
        ret+=v1[i]*v2[i];
    }

    return ret;
}


std::vector<double>
operator*(SparseMatrix &m, const std::vector<double> &v){
    std::vector<double> ret(m.numRows());
    for(std::map<std::pair<int,int>,double>::const_iterator iterator = m.getData().begin(); iterator != m.getData().end(); iterator++) {
        ret[(iterator->first).first] += m((iterator->first).first,(iterator->first).second)*v[(iterator->first).second];
    }
    return ret;
}

std::ostream& 
operator<< (std::ostream& out, const std::vector<double> & v){
    for(size_t i = 0; i < v.size(); ++i){
        out << v[i] << std::endl;
    }
    
    return out;
}

#endif //OPERATORS_H